
import sys
sys.path.append("../")

import pyskynet

remote_addr = "127.0.0.1:8081"
pyskynet.start()

addr = pyskynet.scriptservice("""
local pyskynet = require "pyskynet"
local cluster = require "pyskynet.cluster"
local foreign = require "pyskynet.foreign"
local numsky = require "numsky"

foreign.dispatch("getbuf", function(a)
    return "jklrewklrewjklrw0"
end)

foreign.dispatch("getarr", function(a)
    local arr = numsky.zeros({3,4})
    for i=1,3 do
        for j=1,4 do
            arr[{i,j}] = i + j
        end
    end
    return arr
end)

pyskynet.start(function()
    cluster.open(8081)
end)
""")

pyskynet.scriptservice("""
local pyskynet = require "pyskynet"
local cluster = require "pyskynet.cluster"
local foreign = require "pyskynet.foreign"

local addr = tonumber(...)
pyskynet.start(function()
    local remote_addr = "127.0.0.1:8081"
    local a = cluster.call(remote_addr, addr, "getbuf")
    print(a)
    local b = cluster.call(remote_addr, addr, "getarr")
    print(b)
end)
""", str(addr))

import pyskynet.cluster

buf, = pyskynet.cluster.call(remote_addr, addr, "getbuf")
print(buf)

arr, = pyskynet.cluster.call(remote_addr, addr, "getarr")
print(arr)

#pyskynet.join()




import sys
sys.path.append("../")

import pyskynet
from pyskynet import setenv, getenv


def test_old_env():
    print(getenv(b"rewrw"))
    setenv(b"rewrw", b"rewrewrewrewrwe")
    setenv(b"rew", {"rew":321})
    pyskynet.start()
    pyskynet.scriptservice("""
        local pyskynet = require "pyskynet"

        pyskynet.start(function()
            print(pyskynet.getenv("rewrw"))
            print(pyskynet.getenv("rew").rew)
        end)
    """)




test_old_env()





import sys
sys.path.append("../")

import pyskynet.foreign as foreign
import pyskynet.skynet_py_foreign_seri as foreign_seri

import numpy as np

def equal(left, right):
    if type(left) != type(right):
        return (left == right) == True
    elif type(left)==memoryview:
        return left.tobytes() == right.tobytes()
    elif type(left) == dict:
        if len(left) != len(right):
            return False
        for k in left:
            if not (k in right):
                return False
            if not equal(left[k], right[k]):
                return False
        return True
    elif type(left) == np.ndarray:
        return left.shape == right.shape and (left == right).all()
    else:
        return left == right


def check_case(*args):
    for name, pack, unpack in [
            ("foreign", foreign_seri.pack, foreign_seri.unpack),
            ("foreign_remote", foreign_seri.remotepack, foreign_seri.remoteunpack)]:
        # pack to capsule
        msg_ptr, msg_size = pack(*args)
        result = unpack(msg_ptr, msg_size)
        for i in range(len(args)):
            assert equal(args[i], result[i]), "check error :%s,%s,%s,%s,%s"%(name,
                    args[i],  type(args[i]),
                    result[i], type(result[i]))
        # pack to bytes
        msg_ptr, msg_size = pack(*args)
        msg_bytes = foreign_seri.tobytes(msg_ptr, msg_size)
        result = unpack(msg_bytes)
        for i in range(len(args)):
            assert equal(args[i], result[i]), "check error :%s,%s,%s,%s,%s"%(name,
                    args[i],  type(args[i]),
                    result[i], type(result[i]))
    print("check ok:", *args)

def test_number():
    t = (0, 0x10, 0x100, 0x1000, 0x10000, 0x100000, -4)
    check_case(*t)
    t = (1, 0x20, 0x200, 0x2000, 0x20000, 0x200000, -8)
    check_case(*t)

def test_dict():
    t = (1,{1:123,b"bfds":321,b"rewrw":{b"rewrw":b"rewrewrwrw"}}, 32132131321312)
    check_case(*t)
    t = ({1:123,b"bfds":321,b"rewrw":{b"rewrw":b"rewrewrwrw"}},)
    check_case(*t)

def test_list():
    t = ([1,2,3,4], [i for i in range(100)], [{321:b"rewrw"}, {b"fdsfds":234234}])
    check_case(*t)


def test_arr():
    t1 = np.array([0,1,2,3], dtype=np.bool)
    t2 = np.array([0,1,2,3], dtype=np.int8)
    t3 = np.array([0,1,2,3], dtype=np.uint8)
    t4 = np.array([0,1,2,3], dtype=np.int16)
    check_case(t1,t2,t3,t4)
    t1 = np.array([0,1,2,3], dtype=np.uint16)
    t2 = np.array([0,1,2,3], dtype=np.int32)
    t3 = np.array([0,1,2,3], dtype=np.uint32)
    t4 = np.array([0,1,2,3], dtype=np.int64)
    check_case(t1,t2,t3,t4)
    t1 = np.array([[1,2,3,4], [4,5,6,7]])
    t2 = np.array([[1,2,3,4], [4,5,6,7], [4,5,6,7]])
    t3 = np.array([[[1,2,3],[3,4,5]], [[3,4,5],[5,6,7]]])
    check_case(t1,t2,t3)

test_number()
test_dict()
test_arr()
test_list()



# cython: language_level=3

from cpython.pycapsule cimport PyCapsule_New, PyCapsule_GetPointer, PyCapsule_SetPointer, PyCapsule_GetName
from cpython.pycapsule cimport PyCapsule_CheckExact
from cpython.unicode cimport PyUnicode_CheckExact, PyUnicode_AsUTF8String
from cpython.bytes cimport PyBytes_CheckExact, PyBytes_AS_STRING
from libc.string cimport memcpy, strcmp
from cpython.ref cimport Py_XDECREF, PyObject

from skynet_py cimport *


cdef void free_pyptr(object capsule):
    cdef void *ptr = PyCapsule_GetPointer(capsule, "pyptr")
    skynet_free(ptr)

def crecv():
    cdef SkynetPyMessage msg
    cdef int ret = skynet_py_queue_pop(&msg)
    while ret == 0 and msg.type == PTYPE_DECREF_PYTHON:
        Py_XDECREF(<PyObject*>msg.data)
        ret = skynet_py_queue_pop(&msg)
    if ret != 0:
        return None, None, None, None, None
    else:
        # TODO when msg.type is error , msg.data will be nil ?
        if msg.data != NULL:
            return msg.source, msg.type, msg.session, PyCapsule_New(msg.data, "pyptr", free_pyptr), msg.size
        else:
            return msg.source, msg.type, msg.session, b"", 0

# see lsend in lua-skynet.c
def csend(py_dst, int type_id, py_session, py_msg, py_size=None):
    assert skynet_py_address() > 0, "skynet threads has not been started yet, call 'pyskynet.start()' first."
    # 1. check dst
    cdef char * dstname = NULL
    cdef int dst = 0
    cdef bytes py_dst_bytes
    if PyBytes_CheckExact(py_dst):
        dstname = py_dst
    elif PyUnicode_CheckExact(py_dst):
        py_dst_bytes = PyUnicode_AsUTF8String(py_dst)
        dstname = py_dst_bytes
    else:
        dst = py_dst
    # 2. check session
    cdef int session = 0
    if py_session is None:
        type_id |= PTYPE_TAG_ALLOCSESSION
        session = 0
    else:
        session = py_session
    # 3. check ptr, size
    cdef char * ptr = NULL
    cdef int size = 0
    cdef bytes py_msg_bytes
    if PyCapsule_CheckExact(py_msg):
        type_id |= PTYPE_TAG_DONTCOPY
        ptr = <char *>PyCapsule_GetPointer(py_msg, "cptr")
        size = py_size
    elif PyBytes_CheckExact(py_msg):
        ptr = py_msg
        size = len(py_msg)
    elif PyUnicode_CheckExact(py_msg):
        py_msg_bytes = PyUnicode_AsUTF8String(py_msg)
    else:
        raise Exception("type:%s unexcept when skynet csend"%type(py_msg))

    if dstname == NULL:
        session = skynet_py_send(dst, type_id, session, ptr, size)
    else:
        session = skynet_py_sendname(dstname, type_id, session, ptr, size)
    if session < 0:
        raise Exception("skynet send exception")
    else:
        return session

# extern
def tostring(capsule, size_t size):
    b = tobytes(capsule, size)
    return b.decode()

# extern
def tobytes(capsule, size_t size):
    cdef char *name = PyCapsule_GetName(capsule)
    cdef char *ptr = <char *>PyCapsule_GetPointer(capsule, name)
    cdef bytes s
    if strcmp(name, "cptr") == 0 or strcmp(name, "pyptr") == 0:
        s = ptr[:size]
        return s
    else:
        raise Exception("capsule unpack failed for name = %s")

class PTYPEEnum(object):
    def __init__(self):
        self.PTYPE_TEXT=PTYPE_TEXT
        self.PTYPE_RESPONSE=PTYPE_RESPONSE
        self.PTYPE_MULTICAST=PTYPE_MULTICAST
        self.PTYPE_CLIENT=PTYPE_CLIENT
        self.PTYPE_SYSTEM=PTYPE_SYSTEM
        self.PTYPE_HARBOR=PTYPE_HARBOR
        self.PTYPE_SOCKET=PTYPE_SOCKET
        self.PTYPE_ERROR=PTYPE_ERROR
        self.PTYPE_QUEUE=PTYPE_RESERVED_QUEUE
        self.PTYPE_DEBUG=PTYPE_RESERVED_DEBUG
        self.PTYPE_LUA=PTYPE_RESERVED_LUA
        self.PTYPE_SNAX=PTYPE_RESERVED_SNAX
        self.PTYPE_TRACE=12 # TRACE defined in skynet.lua
        self.PTYPE_FOREIGN=PTYPE_FOREIGN
        self.PTYPE_FOREIGN_REMOTE=PTYPE_FOREIGN_REMOTE
        self.PTYPE_TAG_ALLOCSESSION=PTYPE_TAG_ALLOCSESSION
        self.PTYPE_TAG_DONTCOPY=PTYPE_TAG_DONTCOPY


SKYNET_PTYPE_id_to_name = {}
SKYNET_PTYPE_name_to_id = {}
SKYNET_PTYPE = PTYPEEnum()

for key, id in SKYNET_PTYPE.__dict__.items():
    name = key[6:].lower()
    SKYNET_PTYPE_id_to_name[id] = name
    SKYNET_PTYPE_name_to_id[name] = id
    locals()[key] = id


SKYNET_PTYPE_user_builtin_ids = [SKYNET_PTYPE.PTYPE_LUA, SKYNET_PTYPE.PTYPE_CLIENT, SKYNET_PTYPE.PTYPE_SOCKET, SKYNET_PTYPE.PTYPE_TEXT, SKYNET_PTYPE.PTYPE_FOREIGN, SKYNET_PTYPE.PTYPE_FOREIGN_REMOTE]
def user_assert_id_name(id, name):
    is_builtin = False
    if name in SKYNET_PTYPE_name_to_id:
        assert id == SKYNET_PTYPE_name_to_id[name], "skynet proto type name=%s must bind id=%s" % (name, id)
        is_builtin = True
    if id in SKYNET_PTYPE_id_to_name:
        assert name == SKYNET_PTYPE_id_to_name[id], "skynet proto type id=%s must bind name=%s" % (id, name)
        is_builtin = True
    if is_builtin:
        assert id in SKYNET_PTYPE_user_builtin_ids, "skynet proto can only register builtin proto 'lua' or 'client' or 'socket' or 'text' or 'foreign' or 'foreign_remote', but get name=%s" % name

    assert 0 <= id and id <= 255, "skynet proto type id must be less than 256 but get id=%s" % id




import sys
sys.path.append("../")

import pyskynet
import pyskynet.foreign as foreign

import numpy as np

pyskynet.start()


service = pyskynet.scriptservice("""
        local pyskynet = require "pyskynet"
        local foreign = require "pyskynet.foreign"
        local ns = require "numsky"
        foreign.dispatch("dosth", function(arr)
            return ns.array({1,2,3,4})
        end)
        pyskynet.start(function()
        end)
""")


arr = np.zeros(10000)
arr1, = foreign.call(service, "dosth", arr)

print(arr1, arr1.base)


pyskynet.join()




# 1. import gevent, gevent must use libuv
import gevent
gevent.config.loop = "libuv"
import os
import sys
import gevent.libuv._corecffi as libuv_cffi
import gevent.event

# 2. dlopen with flags RTLD_GLOBAL
flags = sys.getdlopenflags()
sys.setdlopenflags(flags | os.RTLD_GLOBAL)
import pyskynet.skynet_py_main as skynet_py_main
sys.setdlopenflags(flags)

# 3. some module
import pyskynet.proto as pyskynet_proto
import pyskynet.skynet_py_mq as skynet_py_mq

__boot_event = gevent.event.Event()
__exit_event = gevent.event.Event()
__watcher = gevent.get_hub().loop.async_()

boot_service = None

__init_funcs = []


# first callback, waiting for skynet_py_boot
def __first_msg_callback():
    global boot_service
    import pyskynet.skynet_py_foreign_seri
    source, type_id, session, ptr, length = skynet_py_mq.crecv()
    # assert first message ( c.send(".python", 0, 0, "") )
    assert type_id == 0, "first message type must be 0 but get %s" % type_id
    assert session == 0, "first message session must be 0 but get %s" % session
    boot_service, = pyskynet.skynet_py_foreign_seri.unpack(ptr, length)
    __watcher.callback = lambda: gevent.spawn(pyskynet_proto.async_handle)
    gevent.spawn(pyskynet_proto.async_handle)
    __boot_event.set()


# preinit, register libuv items
def __preinit():
    __watcher.start(__first_msg_callback)
    p_uv_async_send = libuv_cffi.ffi.addressof(libuv_cffi.lib, "uv_async_send")
    p_uv_async_t = __watcher._watcher
    skynet_py_main.init(libuv_cffi.ffi, p_uv_async_send, p_uv_async_t)


__preinit()

SKYNET_ROOT = os.path.join(os.path.abspath(
    os.path.dirname(__file__)), "../skynet")
PYSKYNET_ROOT = os.path.abspath(os.path.dirname(__file__))

boot_config = {
    "thread": 1,

    # skynet service path
    "cservice": [SKYNET_ROOT+"/cservice/?.so", PYSKYNET_ROOT+"/service/?.so"],
    "luaservice": [SKYNET_ROOT+"/service/?.lua", PYSKYNET_ROOT+"/service/?.lua", "./?.lua"],

    # lua require path
    "lua_cpath": [SKYNET_ROOT+"/luaclib/?.so", PYSKYNET_ROOT+"/lualib/?.so", "./?.so"],
    "lua_path": [SKYNET_ROOT+"/lualib/?.lua", PYSKYNET_ROOT+"/lualib/?.lua", "./?.lua"],

    # script
    "lualoader": SKYNET_ROOT+"/lualib/loader.lua",
    "bootstrap": "snlua skynet_py_boot",
    "logservice": "snlua",
    "logger": "skynet_py_logger",

    # profile
    "profile": 0,

    # immutable setting, don't use this ...
    "standalone": "1",  # used by service_mgr.lua
    "harbor": "0",  # used by cdummy
}


def init(func):
    if __init_funcs is None:
        func()
    else:
        __init_funcs.append(func)


def start(thread=2, path=[], cpath=[]):
    """
    example:
        start(thread=1,path=["./?.lua"],cpath=["./?.so"])
    """
    global __init_funcs
    assert type(path) == list, "start path must be list"
    for f in path:
        for key in ["lua_path", "luaservice"]:
            boot_config[key].append(f)
    assert type(cpath) == list, "start cpath must be list"
    for f in cpath:
        for key in ["lua_cpath", "cservice"]:
            boot_config[key].append(f)
    boot_config["thread"] = thread
    import pyskynet
    for k, v in boot_config.items():
        if type(v) == list:
            v = ";".join(v)
        pyskynet.setenv(k, v)
    skynet_py_main.start(thread=thread, profile=boot_config["profile"])
    __boot_event.wait()
    funcs = __init_funcs
    __init_funcs = None
    for f in funcs:
        f()


def join():
    __exit_event.wait()


__python_exit = exit


def exit():
    skynet_py_main.exit()
    __exit_event.set()
    __python_exit()


def main():
    import argparse
    parser = argparse.ArgumentParser(description='pyskynet fast entry')
    parser.add_argument("script", type=str,
                        help="lua service script file", nargs='?', default="")
    parser.add_argument("args", type=str,
                        help="service arguments", nargs='*', default="")
    args = parser.parse_args()
    if args.script != "":
        import pyskynet
        start()
        try:
            with open(args.script) as fo:
                script = fo.read()
            pyskynet.scriptservice([script, args.script], *args.args)
            join()
        except pyskynet.proto.PySkynetCallException as err:
            print(err)
            gevent.sleep(1)
        except KeyboardInterrupt:
            return
    else:
        import pyskynet
        import pyskynet.foreign
        start()
        import code
        import sys
        import readline
        readline.parse_and_bind('tab: complete')

        class PySkynet(code.InteractiveConsole):
            def __init__(self, *args, **kwargs):
                super().__init__()
                sys.ps1 = "(lua)> "

            def runsource(self, *args, **kwargs):
                pyskynet.foreign.call(boot_service, "repl", args[0])
                return False

            def raw_input(self, *args, **kwargs):
                try:
                    re = super().raw_input(*args, **kwargs)
                    return re
                except KeyboardInterrupt:
                    pass
                print("")
                exit()
        PySkynet().interact()




import sys
sys.path.append("../")

import pyskynet
import pyskynet.cluster as cluster
import pyskynet.foreign as foreign
import numpy as np
import time

@foreign.dispatch("dosth")
def dosth(a):
    print(a)
    return a

pyskynet.start()


cluster.open(8080)



a1 = np.arange(100)
t1 = time.time()
a2, = cluster.call("127.0.0.1:8080", ".python", "dosth", a1)
t2 = time.time()
print(t2-t1)

t1 = time.time()
a2, = cluster.call("127.0.0.1:8080", ".python", "dosth", a1)
t2 = time.time()
print(t2-t1)

pyskynet.join()




import sys
sys.path.append("../")


import cv2
import pyskynet
import pyskynet.foreign as foreign

pyskynet.start()

service = pyskynet.scriptservice("""
        local pyskynet = require "pyskynet"
        local foreign = require "pyskynet.foreign"
        local ns = require "numsky"
        local tinygl = require "numsky.tinygl"
        local rapidjson = require "rapidjson"


        local box_v = ns.array({{-4.9,-4.9,-3}, {-4.9,-4.9,-3}, {-4.9,-4.9,-3}}, ns.float32)
        local box_i = ns.array({{1,2,3}}, ns.int32)
        local mesh1 = tinygl.mesh(box_v, box_i)

        local box_v = ns.array({{4.9,4.9,-3}, {4.9,4.9,-3}, {4.9,4.9,-3}}, ns.float32)
        local box_i = ns.array({{1,2,3}}, ns.int32)
        local mesh2 = tinygl.mesh(box_v, box_i)

        local camera = tinygl.camera(101,101,3)
        camera:rotation(0,0,0)
        camera:position(0,0,0)
        camera:ortho(-10,10,-10,10)
        local rect1 = tinygl.rect(-0.5,-0.5,2,1)
        rect1:rotation(0,0,0)
        rect1:scale(5,5,0)
        local rect2 = tinygl.rect(-0.5,-0.5,1,1)
        rect2:rotation(0,0,0)
        rect2:scale(5,5,0)
        rect2:setColor(255,255,255)
        local circle = tinygl.circle(0,0,5)
        circle:setColor(0,255,255,255)
        circle:setFillType(2)
        circle:rotation(0,0,90)
        circle:scale(1,0.5,1)
        circle:position(0,0,0)
        local tri = tinygl.circle(0,0,5,3)
        tri:setColor(255,255,255)
        local sector = tinygl.sector(0,0,5,15)
        sector:setColor(255,255,0)
        local line = tinygl.line(0,0,5,5)
        line:setColor(255,255,0)
        local point = tinygl.point(-3,-3)
        point:setColor(255,255,0)
        foreign.dispatch("test", function()
            camera:draw(rect1)
            camera:draw(rect2)
            camera:draw(sector)
            camera:draw(line)
            camera:draw(point)
            camera:draw(mesh1)
            camera:draw(mesh2)
            local pixel_arr, depth_arr = camera:array()
            return pixel_arr:copy(), depth_arr:copy()
        end)
        pyskynet.start(function()
        end)
""")

arr1, arr2 = foreign.call(service, "test")

import numpy as np



cv2.imwrite("tinygl1.bmp", arr1);
cv2.imwrite("tinygl2.bmp", arr2);

#arr1 = arr1.reshape((arr1.shape[0], arr1.shape[1]))
#print(np.where(arr1==255))
#print(arr1)
#print(np.sum(arr1))
print(arr1.shape)




import pyskynet.skynet_py_foreign_seri as foreign_seri
import pyskynet.proto as pyskynet_proto
import pyskynet.skynet_py_mq


PTYPE_FOREIGN = pyskynet.skynet_py_mq.PTYPE_FOREIGN
PTYPE_FOREIGN_REMOTE = pyskynet.skynet_py_mq.PTYPE_FOREIGN_REMOTE


class CMDDispatcher(object):
    def __init__(self):
        self.cmd2func = {}

    def __call__(self, first, *args):
        return self.cmd2func[first](*args)

    def __setitem__(self, k, v):
        if isinstance(k, str):
            self.cmd2func[k] = v
            self.cmd2func[k.encode("ascii")] = v
        elif isinstance(k, bytes):
            self.cmd2func[k] = v
            self.cmd2func[k.decode("ascii")] = v
        else:
            self.cmd2func[k] = v

    def __getitem__(self, k):
        return self.cmd2func[k]


CMD = CMDDispatcher()

pack = foreign_seri.pack
unpack = foreign_seri.unpack
remotepack = foreign_seri.remotepack
remoteunpack = foreign_seri.remoteunpack


def __foreign_dispatch(session, source, argtuple):
    if session != 0:
        pyskynet_proto.ret(*foreign_seri.pack(CMD(*argtuple)))
    else:
        CMD(*argtuple)


def __foreign_remote_dispatch(session, source, argtuple):
    if session != 0:
        pyskynet_proto.ret(*foreign_seri.remotepack(CMD(*argtuple)))
    else:
        CMD(*argtuple)


# dispatch foreign message
pyskynet_proto.register_protocol(
        id=PTYPE_FOREIGN,
        name="foreign",
        pack=foreign_seri.pack,
        unpack=foreign_seri.unpack,
        dispatch=__foreign_dispatch,
        )

# dispatch foreign message
pyskynet_proto.register_protocol(
        id=PTYPE_FOREIGN_REMOTE,
        name="foreign_remote",
        pack=foreign_seri.remotepack,
        unpack=foreign_seri.remoteunpack,
        dispatch=__foreign_remote_dispatch,
        )


def dispatch(cmd=None, func=None):
    global CMD

    def wrapper(func):
        CMD[cmd] = func
        return func
    if not (func is None):
        assert callable(func), "dispatch's second arg must be callable"
        CMD[cmd] = func
    elif callable(cmd):
        CMD = func
    elif isinstance(cmd, dict):
        for k, v in cmd.items():
            CMD[k] = v
    elif isinstance(cmd, str) or isinstance(cmd, bytes):
        return wrapper
    else:
        raise Exception("dispatch failed for unexception args")


def call(addr, *args):
    return foreign_seri.unpack(*pyskynet_proto.rawcall(
        addr, PTYPE_FOREIGN, *foreign_seri.pack(*args)))


def send(addr, *args):
    pyskynet_proto.rawsend(addr, PTYPE_FOREIGN, *foreign_seri.pack(*args))



# cython: language_level=3

import _cffi_backend
from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_New, PyCapsule_CheckExact
from cpython.bytes cimport PyBytes_FromString, PyBytes_FromStringAndSize, PyBytes_CheckExact
from libc.stdio cimport sscanf, sprintf

from skynet_py cimport *

cdef extern from "skynet_env.h":
    const char * skynet_getenv(const char *key);

cdef extern from "skynet_modify/skynet_py.h":
    void *skynet_py_setlenv(const char *key, const char *value_str, size_t sz)
    const char *skynet_py_getlenv(const char *key, size_t *sz);
    const char *skynet_py_nextenv(const char *key)

ctypedef (char *)(* f_type)(object, object)

def init(ffi, async_send, async_handle):
    # because gevent's libuv bind with cffi, so we can get pointer by this way
    cdef void ** _cffi_exports = <void **>PyCapsule_GetPointer(_cffi_backend._C_API, "cffi")
    # get the '_cffi_to_c_pointer' function in _cffi_include.h of cffi
    cdef f_type _cffi_to_c_pointer = <f_type>_cffi_exports[11]
    skynet_py_init(
            <int (*)(void*)>_cffi_to_c_pointer(async_send, ffi.typeof(async_send)),
            <void *>_cffi_to_c_pointer(async_handle, ffi.typeof(async_handle))
            )

cdef __check_bytes(s):
    t = type(s)
    if t == str:
        return s.encode("utf-8")
    elif t == bytes:
        return s
    else:
        raise Exception("type %s can't convert to bytes" % str(t))

def setlenv(key, capsule_or_bytes, py_sz=None):
    cdef size_t sz
    cdef const char *ptr
    if PyCapsule_CheckExact(capsule_or_bytes):
        ptr = <char *>PyCapsule_GetPointer(capsule_or_bytes, "cptr")
        sz = py_sz
    elif PyBytes_CheckExact(capsule_or_bytes):
        ptr = capsule_or_bytes
        sz = len(capsule_or_bytes)
    else:
        raise Exception("skynet_py env value must be bytes or pointer")
    if not (key is None):
        key = __check_bytes(key)
        skynet_py_setlenv(key, ptr, sz)
        return None
    cdef void *newkey = skynet_py_setlenv(NULL, ptr, sz)
    cdef char addr[32];
    cdef int k = sprintf(addr, "%p", newkey)
    return addr[:k]


def getlenv(key):
    key = __check_bytes(key)
    cdef size_t sz
    cdef const char * value = skynet_py_getlenv(key, &sz);
    if value != NULL:
        return PyBytes_FromStringAndSize(value, sz);
    else:
        return None

def nextenv(key):
    cdef const char * ptr
    if key is None:
        ptr = skynet_py_nextenv(NULL)
    else:
        key = __check_bytes(key)
        ptr = skynet_py_nextenv(key)
    if ptr == NULL:
        return None
    else:
        return PyBytes_FromString(ptr)

def start(int thread, int profile):
    cdef skynet_config config;
    config.thread = thread
    config.profile = profile
    # use getenv for a stable ptr
    config.module_path = skynet_getenv("cservice")
    config.bootstrap = skynet_getenv("bootstrap")
    config.logservice = skynet_getenv("logservice")
    config.logger = skynet_getenv("logger")
    # ignore
    config.harbor = 0
    config.daemon = NULL # just ignore daemon
    skynet_py_start(&config)

def exit():
    skynet_py_exit()

def self():
    return skynet_py_address()




import pyskynet.skynet_py_mq as skynet_py_mq
import pyskynet.skynet_py_foreign_seri as foreign_seri
import gevent
from gevent.event import AsyncResult

SKYNET_PTYPE = skynet_py_mq.SKYNET_PTYPE

#####################
# py<->skynet proto #
#####################

local_session_to_ar = {}

pyskynet_proto_dict = {}


class PySkynetProto(object):
    def __init__(self, id, name, pack=None, unpack=None, dispatch=None):
        self.id = id
        self.name = name
        self.pack = pack
        self.unpack = unpack
        self.dispatch = dispatch


# skynet.lua
def register_protocol(id, name, pack=None, unpack=None, dispatch=None):
    if type(name) == bytes:
        name = name.decode()
    assert id not in pyskynet_proto_dict, "pyskynet proto id=%s existed" % id
    assert name not in pyskynet_proto_dict, "pyskynet proto name=%s existed" % name
    assert type(name) == str and type(id) == int and \
    id >= 0 and id <= 255, "pyskynet proto register failed id=%s, name=%s" % (id, name)
    psproto = PySkynetProto(id, name, pack, unpack, dispatch)
    pyskynet_proto_dict[id] = psproto
    pyskynet_proto_dict[name] = psproto


# skynet.lua
def _error_dispatch(error_session, error_source, arg):
    co = gevent.getcurrent()
    co_to_remote_session.pop(co)
    co_to_remote_address.pop(co)
    local_session_to_ar[error_session].set(None)


# skynet.lua
register_protocol(
        id=SKYNET_PTYPE.PTYPE_LUA,
        name="lua",
        pack=foreign_seri.luapack,
        unpack=foreign_seri.luaunpack,
        )

# skynet.lua
register_protocol(
        id=SKYNET_PTYPE.PTYPE_ERROR,
        name="error",
        pack=foreign_seri.luapack,
        unpack=lambda a, b: None,
        dispatch=_error_dispatch,
        )


# skynet.lua
def dispatch(name, func):
    """
        dispatch in skynet.lua,
        set dispatch
    """
    pyskynet_proto_dict[name].dispatch = func


####################
# code for session #
####################

co_to_remote_session = {}
co_to_remote_address = {}


class PySkynetCallException(Exception):
    pass


# skynet.lua
def rawcall(dst, type_name_or_id, msg_ptr, msg_size):
    """
        rawcall in skynet.lua, rpc call
    """
    psproto = pyskynet_proto_dict[type_name_or_id]
    session = skynet_py_mq.csend(dst, psproto.id, None, msg_ptr, msg_size)
    ar = AsyncResult()
    local_session_to_ar[session] = ar
    re = ar.get()
    if re:
        return re
    else:
        raise PySkynetCallException("call failed from %s" % dst)


# skynet.lua
def rawsend(dst, type_name_or_id, msg_ptr, msg_size):
    """
        rawsend in skynet.lua, send don't need ret
    """
    psproto = pyskynet_proto_dict[type_name_or_id]
    skynet_py_mq.csend(dst, psproto.id, 0, msg_ptr, msg_size)


# skynet.lua
def call(addr, type_name_or_id, *args):
    return foreign_seri.unpack(*rawcall(addr, type_name_or_id, *foreign_seri.pack(*args)))


# skynet.lua
def send(addr, type_name_or_id, *args):
    rawsend(addr, type_name_or_id, *foreign_seri.pack(*args))


# skynet.lua
def ret(ret_msg_ptr, ret_size):
    """
        ret in skynet.lua, return for other's call
    """
    co = gevent.getcurrent()
    session = co_to_remote_session.pop(co)
    source = co_to_remote_address.pop(co)
    if session == 0:
        pass  # send don't need ret
    else:
        skynet_py_mq.csend(source, SKYNET_PTYPE.PTYPE_RESPONSE, session, ret_msg_ptr, ret_size)
    # TODO if package is to large, c++ will trigger some exception...


################
# raw dispatch #
################
def async_handle():
    """
        python actor's main loop, recv and deal message
    """
    source, type_id, session, ptr, length = skynet_py_mq.crecv()
    if source is None:
        return
    else:
        gevent.spawn(async_handle)
    if type_id == SKYNET_PTYPE.PTYPE_RESPONSE:
        # TODO exception
        ar = local_session_to_ar.pop(session)
        ar.set((ptr, length))
    else:
        # TODO exception
        psproto = pyskynet_proto_dict[type_id]
        co = gevent.getcurrent()
        co_to_remote_session[co] = session
        co_to_remote_address[co] = source
        psproto.dispatch(session, source, psproto.unpack(ptr, length))
        if co in co_to_remote_session:
            if session != 0:
                print("Maybe forgot response session %s from %s " % (session, source))
            co_to_remote_session.pop(co)
            co_to_remote_address.pop(co)



# cython: language_level=3

from libc.string cimport memcpy, strcmp
from cpython.pycapsule cimport PyCapsule_New, PyCapsule_GetPointer, PyCapsule_GetName
# import some type check
from cpython.pycapsule cimport PyCapsule_CheckExact
from cpython.unicode cimport PyUnicode_CheckExact, PyUnicode_AsUTF8String, PyUnicode_DecodeUTF8
from cpython.bytes cimport PyBytes_CheckExact, PyBytes_AS_STRING, PyBytes_FromStringAndSize, PyBytes_FromString, PyBytes_GET_SIZE
from cpython.object cimport PyObject, PyTypeObject
from cpython.memoryview cimport PyMemoryView_Check
from cpython.bytearray cimport PyByteArray_AS_STRING
from numpy cimport PyArray_CheckExact
import numpy as np
cimport numpy as cnp
cnp.import_array()

from skynet_py cimport *

DEF BLOCK_SIZE = 128

cdef extern from "skynet_py_foreign_seri_ext.c": #from "lua-foreign_seri.c":
    cdef enum:
        MODE_LUA
        MODE_FOREIGN
        MODE_FOREIGN_REMOTE

    cdef struct block:
        block* next
        char *buffer
    # for read
    cdef struct read_block:
        pass
    cdef struct foreign_read_block:
        int mode
    void foreign_rball_init(foreign_read_block* rb, char* buffer, int size, int mode)
    read_block* rb_cast(foreign_read_block* rb)
    void* foreign_rb_read(foreign_read_block* rb, int sz)
    uint8_t COMBINE_TYPE(uint8_t, uint8_t)

    # for write
    cdef struct write_block:
        int len
    cdef struct foreign_write_block:
        write_block wb
        int mode
    void foreign_wb_init(foreign_write_block* wb, block* b, int mode)
    write_block* wb_cast(foreign_write_block* wb);
    void wb_free(write_block *wb)
    void wb_nil(write_block* wb)
    void wb_integer(write_block* wb, lua_Integer v)
    void wb_real(write_block* wb, double v)
    void wb_boolean(write_block* wb, int v)
    void wb_pointer(write_block *wb, void *v)
    void wb_string(write_block *wb, const char *ptr, int sz)
    void foreign_wb_write(foreign_write_block *wb, const void *buf, int sz)
    cdef enum:
        TYPE_NIL
        TYPE_BOOLEAN
        TYPE_NUMBER
        TYPE_NUMBER_ZERO
        TYPE_NUMBER_BYTE
        TYPE_NUMBER_WORD
        TYPE_NUMBER_DWORD
        TYPE_NUMBER_QWORD
        TYPE_NUMBER_REAL

        TYPE_USERDATA
        TYPE_SHORT_STRING
        TYPE_LONG_STRING
        TYPE_TABLE
        TYPE_FOREIGN_USERDATA
    cdef enum:
        MAX_COOKIE
        MAX_DEPTH

    # deal for PyArray & PyMemoryView
    bint PyArray_foreign_check_typechar(object py_obj)
    void wb_foreign_PyArray(foreign_write_block *wb, object arr, object arr_iter)
    object unpack_PyArray(foreign_read_block *wb, int cookie)

########################
# functions for unpack #
########################

cdef double py_get_real(foreign_read_block* rb) except *:
    cdef double n_double
    cdef void * ptr
    ptr = foreign_rb_read(rb, sizeof(n_double))
    if(ptr==NULL):
        raise Exception("invalid stream")
    memcpy(&n_double, ptr, sizeof(n_double));
    return n_double

cdef lua_Integer py_get_integer(foreign_read_block* rb, int cookie) except *:
    cdef uint8_t n_uint8
    cdef uint16_t n_uint16
    cdef int32_t n_int32
    cdef int64_t n_int64
    cdef void * ptr
    if(cookie==TYPE_NUMBER_ZERO):
        return 0
    elif(cookie==TYPE_NUMBER_BYTE):
        ptr = foreign_rb_read(rb, sizeof(n_uint8));
        if (ptr == NULL):
            raise Exception("invalid stream")
        memcpy(&n_uint8, ptr, sizeof(n_uint8));
        return n_uint8
    elif(cookie==TYPE_NUMBER_WORD):
        ptr = foreign_rb_read(rb, sizeof(n_uint16));
        if(ptr == NULL):
            raise Exception("invalid stream")
        memcpy(&n_uint16, ptr, sizeof(n_uint16));
        return n_uint16
    elif(cookie==TYPE_NUMBER_DWORD):
        ptr = foreign_rb_read(rb, sizeof(n_int32));
        if(ptr == NULL):
            raise Exception("invalid stream")
        memcpy(&n_int32, ptr, sizeof(n_int32));
        return n_int32
    elif(cookie==TYPE_NUMBER_QWORD):
        ptr = foreign_rb_read(rb, sizeof(n_int64));
        if(ptr == NULL):
            raise Exception("invalid stream")
        memcpy(&n_int64, ptr, sizeof(n_int64));
        return n_int64
    else:
        raise Exception("invalid stream")

#cdef skynet_foreign* py_get_foreign(foreign_read_block *rb) except NULL:
#    cdef skynet_foreign* userdata = NULL
#    cdef void * buffer = NULL
#    cdef void ** v
#    cdef int32_t * psize
#    if rb.mode == MODE_FOREIGN:
#        v = <void **>foreign_rb_read(rb,sizeof(userdata));
#        if v == NULL:
#            raise Exception("invalid stream")
#        memcpy(&userdata, v, sizeof(userdata));
#    elif rb.mode == MODE_FOREIGN_REMOTE:
#        userdata = skynet_foreign_deserialize(rb, <void *(*)(void*, int)>foreign_rb_read);
#        if userdata == NULL:
#            raise Exception("invalid stream")
#    else:
#        raise Exception("unexcept mode for TYPE_FOREIGN_USERDATA")
#    return userdata

cdef void* py_get_pointer(foreign_read_block *rb) except NULL:
    cdef void * userdata
    cdef void ** v = <void **>foreign_rb_read(rb,sizeof(userdata));
    if v == NULL:
        raise Exception("invalid stream")
    memcpy(&userdata, v, sizeof(userdata));
    return userdata

cdef char * py_get_string(foreign_read_block *rb, int value_type, int cookie, size_t *out) except NULL:
    cdef char * p = NULL
    cdef uint16_t *plen2
    cdef uint32_t *plen4
    if value_type==TYPE_SHORT_STRING:
        p = <char *>foreign_rb_read(rb, cookie)
        out[0] = cookie
    elif value_type==TYPE_LONG_STRING:
        if(cookie == 2):
            plen2 = <uint16_t *>foreign_rb_read(rb, 2)
            if (plen2 == NULL):
                raise Exception("invalid stream")
            p = <char *>foreign_rb_read(rb, plen2[0])
            out[0] = plen2[0]
        elif(cookie == 4):
            plen4 = <uint32_t *>foreign_rb_read(rb, 4)
            if(plen4 == NULL):
                raise Exception("invalid stream")
            p = <char *>foreign_rb_read(rb, plen4[0])
            out[0] = plen4[0]
    return p

cdef void py_push_value(l, foreign_read_block *rb, int value_type, int cookie, bint iskey) except *:
    cdef char * ptr = NULL
    cdef size_t length = 0
    cdef void * lightuserdata = NULL
    cdef skynet_foreign * foreign_obj = NULL
    if value_type == TYPE_NIL:
        l.append(None)
    elif value_type == TYPE_BOOLEAN:
        l.append(cookie>0)
    elif value_type == TYPE_NUMBER:
        if (cookie == TYPE_NUMBER_REAL):
            l.append(py_get_real(rb))
        else:
            l.append(py_get_integer(rb, cookie))
    elif value_type==TYPE_SHORT_STRING:
        ptr = py_get_string(rb, value_type, cookie, &length)
        #if iskey:
            #l.append(PyUnicode_DecodeUTF8(ptr, length, NULL))
        #else:
        l.append(PyBytes_FromStringAndSize(ptr, length))
    elif value_type==TYPE_LONG_STRING:
        ptr = py_get_string(rb, value_type, cookie, &length)
        #if iskey:
            #l.append(PyUnicode_DecodeUTF8(ptr, length, NULL))
        #else:
        l.append(PyBytes_FromStringAndSize(ptr, length))
    elif value_type==TYPE_TABLE:
        py_unpack_table(l, rb, cookie)
    elif value_type==TYPE_USERDATA:
        lightuserdata = py_get_pointer(rb)
        l.append(PyCapsule_New(lightuserdata, "cptr", NULL))
    elif value_type==TYPE_FOREIGN_USERDATA:
        arr = unpack_PyArray(rb, cookie)
        if arr is None:
            raise Exception("invalid stream when unpacking arr")
        l.append(arr)
    else:
        raise Exception("invalid stream for value type exception")

cdef void py_unpack_one(l, foreign_read_block *rb, bint iskey) except *:
    cdef uint8_t value_type
    cdef uint8_t *value_ptr = <uint8_t*>foreign_rb_read(rb, sizeof(value_type));
    if value_ptr == NULL:
        raise Exception("invalid stream")
    value_type = value_ptr[0]
    py_push_value(l, rb, value_type & 0x7, value_type>>3, iskey);

cdef void py_unpack_table(l, foreign_read_block *rb, int array_size) except *:
    cdef uint8_t value_type
    cdef uint8_t *value_ptr
    cdef int cookie
    if array_size == MAX_COOKIE-1:
        value_ptr = <uint8_t *>foreign_rb_read(rb, sizeof(value_type))
        if value_ptr == NULL:
            raise Exception("invalid stream")
        value_type=value_ptr[0]
        cookie = value_type>>3
        if (value_type & 7) != TYPE_NUMBER or cookie == TYPE_NUMBER_REAL:
            raise Exception("invalid stream")
        array_size = py_get_integer(rb, cookie)
    #l.append(t)
    next_l = []
    for i in range(1, array_size+1):
        py_unpack_one(next_l, rb, 0)
    next_t = {}
    while True:
        py_unpack_one(next_l, rb, 1)
        if next_l[-1] is None:
            next_l.pop()
            break
        py_unpack_one(next_l, rb, 0)
        next_t[next_l[-2]] = next_l[-1]
        next_l.pop()
        next_l.pop()
    if not next_t:
        l.append(next_l)
    else:
        for i, v in enumerate(next_l):
            next_t[i + 1] = v
        l.append(next_t)

cdef void cunpack(l, char *msg, size_t size, int mode) except *:
    cdef foreign_read_block rb
    foreign_rball_init(&rb, msg, size, mode);
    cdef int i = 0
    cdef uint8_t value_type = 0
    cdef uint8_t *value_ptr = NULL
    while True:
        value_ptr = <uint8_t *>foreign_rb_read(&rb, sizeof(value_type))
        if value_ptr == NULL:
            break
        value_type = value_ptr[0]
        py_push_value(l, &rb, value_type & 0x7, value_type>>3, 0);

# extern
cdef py_foreign_unpack(int mode, capsule_or_bytes, py_sz):
    cdef const char *name
    cdef char *ptr
    cdef size_t sz
    l = []
    if PyCapsule_CheckExact(capsule_or_bytes):
        name = PyCapsule_GetName(capsule_or_bytes)
        ptr = <char *>PyCapsule_GetPointer(capsule_or_bytes, name)
        sz = py_sz
        if strcmp(name, "cptr") == 0 or strcmp(name, "pyptr") == 0:
            cunpack(l, ptr, sz, mode)
            return tuple(l)
        else:
            raise Exception("capsule unpack failed for name = %s " % PyBytes_FromString(name))
    elif PyBytes_CheckExact(capsule_or_bytes):
        ptr = <char *>PyBytes_AS_STRING(capsule_or_bytes)
        sz = PyBytes_GET_SIZE(capsule_or_bytes)
        cunpack(l, ptr, sz, mode)
        return tuple(l)
    else:
        raise Exception("Unexcept type %s " % str(type(capsule_or_bytes)))

######################
# functions for pack #
######################
cdef void py_pack_list(foreign_write_block* wb, list_obj, int depth) except *:
    cdef int array_size = len(list_obj)
    cdef uint8_t n
    if array_size >= MAX_COOKIE - 1:
        n = COMBINE_TYPE(TYPE_TABLE, MAX_COOKIE - 1);
        foreign_wb_write(wb, &n, 1);
        wb_integer(wb_cast(wb), array_size)
    else:
        n = COMBINE_TYPE(TYPE_TABLE, array_size);
        foreign_wb_write(wb, &n, 1);
    for v in list_obj:
        py_pack_one(wb,v,depth)
    wb_nil(wb_cast(wb))

cdef void py_pack_dict(foreign_write_block* wb, dict_obj, int depth) except *:
    cdef uint8_t n = COMBINE_TYPE(TYPE_TABLE, 0);
    foreign_wb_write(wb, &n, 1);
    for k, v in dict_obj.items():
        py_pack_one(wb,k,depth)
        py_pack_one(wb,v,depth)
    wb_nil(wb_cast(wb))

cdef void py_pack_one(foreign_write_block* wb, py_arg, int depth) except *:
    cdef int bytes_sz = 0
    cdef char *bytes_ptr = NULL
    cdef const char *name = NULL
    if depth > MAX_DEPTH:
        raise Exception("serialize can't pack too depth table")
    if py_arg is None:
        wb_nil(wb_cast(wb))
    elif PyCapsule_CheckExact(py_arg):
        name = PyCapsule_GetName(py_arg)
        if strcmp(name, "cptr") == 0:
            wb_pointer(wb_cast(wb), PyCapsule_GetPointer(py_arg, "cptr"))
        else:
            raise Exception("unexception capsule")
    elif PyArray_CheckExact(py_arg):
        if not PyArray_foreign_check_typechar(py_arg):
            raise Exception("unexception typechar %s"%py_arg.dtype.char)
        if wb.mode == MODE_FOREIGN:
            wb_foreign_PyArray(wb, py_arg, None)
        elif wb.mode == MODE_FOREIGN_REMOTE:
            arr_iter = py_arg.flat
            wb_foreign_PyArray(wb, py_arg, arr_iter)
        else:
            raise Exception("seri wb foreign unexception mode")
    elif isinstance(py_arg, dict):
        py_pack_dict(wb, py_arg, depth+1)
    elif isinstance(py_arg, list):
        py_pack_list(wb, py_arg, depth+1)
    else:
        py_arg_type = type(py_arg)
        if py_arg_type == int or np.issubdtype(py_arg_type, np.integer):
            wb_integer(wb_cast(wb), py_arg)
        elif py_arg_type == float or np.issubdtype(py_arg_type, np.floating):
            wb_real(wb_cast(wb), py_arg)
        elif py_arg_type == bool:
            wb_boolean(wb_cast(wb), py_arg)
        elif py_arg_type == bytes:
            bytes_sz = len(py_arg)
            bytes_ptr = py_arg
            wb_string(wb_cast(wb), bytes_ptr, bytes_sz)
        elif py_arg_type == bytearray:
            bytes_sz = len(py_arg)
            bytes_ptr = PyByteArray_AS_STRING(py_arg)
            wb_string(wb_cast(wb), bytes_ptr, bytes_sz)
        elif py_arg_type == str:
            py_arg_bytes = PyUnicode_AsUTF8String(py_arg)
            bytes_sz = len(py_arg_bytes)
            bytes_ptr = py_arg_bytes
            wb_string(wb_cast(wb), bytes_ptr, bytes_sz)
        else:
            raise Exception("Unsupport type %s to serialize"%str(py_arg_type))

cdef py_foreign_pack(int mode, argtuple):
    cdef block temp
    temp.next = NULL
    cdef foreign_write_block wb
    foreign_wb_init(&wb, &temp, mode)
    for one in argtuple:
        py_pack_one(&wb, one, 0)
    # seri
    cdef block *b = &temp
    cdef int length = wb.wb.len
    cdef uint8_t* buffer = <uint8_t*> skynet_malloc(length)
    cdef uint8_t* ptr = buffer
    cdef int sz = length
    while(length > 0):
        if length >= BLOCK_SIZE:
            memcpy(ptr, b.buffer, BLOCK_SIZE)
            ptr += BLOCK_SIZE
            length -= BLOCK_SIZE
            b = b.next
        else:
            memcpy(ptr, b.buffer, length)
            break
    wb_free(wb_cast(&wb))
    return PyCapsule_New(buffer, "cptr", NULL), sz

#########################
# outside pack & unpack #
#########################

def tobytes(capsule, size_t size):
    cdef char *ptr = <char *>PyCapsule_GetPointer(capsule, PyCapsule_GetName(capsule))
    return ptr[:size]

def luapack(*args):
    return py_foreign_pack(MODE_LUA, args)

def luaunpack(capsule, size=None):
    return py_foreign_unpack(MODE_LUA, capsule, size)

def pack(*args):
    return py_foreign_pack(MODE_FOREIGN, args)

def unpack(capsule, size=None):
    return py_foreign_unpack(MODE_FOREIGN, capsule, size)

def remotepack(*args):
    return py_foreign_pack(MODE_FOREIGN_REMOTE, args)

def remoteunpack(capsule, size=None):
    return py_foreign_unpack(MODE_FOREIGN_REMOTE, capsule, size)

def trash(capsule, size_t sz):
    cdef void *ptr = PyCapsule_GetPointer(capsule, "cptr")
    skynet_free(ptr)



# cython: language_level=3

cdef extern from "lua.h":
    ctypedef int lua_Integer


cdef extern from "skynet.h":
    ctypedef int int64_t
    ctypedef int int32_t
    ctypedef int uint32_t
    ctypedef int uint16_t
    ctypedef int uint8_t
    void skynet_free(void *)
    void* skynet_malloc(size_t)
    cdef enum:
        PTYPE_TEXT
        PTYPE_RESPONSE
        PTYPE_MULTICAST
        PTYPE_CLIENT
        PTYPE_SYSTEM
        PTYPE_HARBOR
        PTYPE_SOCKET
        PTYPE_ERROR
        PTYPE_RESERVED_QUEUE
        PTYPE_RESERVED_DEBUG
        PTYPE_RESERVED_LUA
        PTYPE_RESERVED_SNAX
        PTYPE_TAG_ALLOCSESSION
        PTYPE_TAG_DONTCOPY

cdef extern from "skynet_foreign/skynet_foreign.h":
    cdef struct skynet_foreign:
        pass
    void skynet_foreign_incref(skynet_foreign *obj);
    void skynet_foreign_decref(skynet_foreign *obj);

cdef extern from "skynet_modify/skynet_py.h":
    cdef struct skynet_config:
        int thread
        int harbor
        int profile
        const char * daemon
        const char * module_path
        const char * bootstrap
        const char * logger
        const char * logservice
    cdef struct SkynetPyMessage:
        int type
        int session
        uint32_t source
        void * data
        size_t size
    cdef enum:
        PTYPE_FOREIGN_REMOTE
        PTYPE_FOREIGN
        PTYPE_DECREF_PYTHON
    int skynet_py_queue_pop(SkynetPyMessage * )
    int skynet_py_send(uint32_t dst, int type, int session, void* msg, size_t sz);
    int skynet_py_sendname(const char *dst, int type, int session, void* msg, size_t sz);
    void skynet_py_init(int (*p_uv_async_send)(void *), void * p_uv_async_t);
    void skynet_py_start(skynet_config * config)
    void skynet_py_exit();
    uint32_t skynet_py_address();





###############################
# some api different from lua #
###############################
import pyskynet.boot
import pyskynet.skynet_py_mq
import pyskynet.foreign
import pyskynet.skynet_py_main as skynet_py_main
import pyskynet.skynet_py_foreign_seri as foreign_seri
import pyskynet.proto as pyskynet_proto

__version__ = '0.0.8'
start = pyskynet.boot.start
join = pyskynet.boot.join
boot_config = pyskynet.boot.boot_config

#############
# proto api #
#############

PTYPE_TEXT = pyskynet.skynet_py_mq.SKYNET_PTYPE.PTYPE_TEXT
PTYPE_CLIENT = pyskynet.skynet_py_mq.SKYNET_PTYPE.PTYPE_CLIENT
PTYPE_SOCKET = pyskynet.skynet_py_mq.SKYNET_PTYPE.PTYPE_SOCKET
PTYPE_LUA = pyskynet.skynet_py_mq.SKYNET_PTYPE.PTYPE_LUA
PTYPE_FOREIGN_REMOTE = pyskynet.skynet_py_mq.SKYNET_PTYPE.PTYPE_FOREIGN_REMOTE
PTYPE_FOREIGN = pyskynet.skynet_py_mq.SKYNET_PTYPE.PTYPE_FOREIGN

pyskynet.rawcall = pyskynet_proto.rawcall
pyskynet.rawsend = pyskynet_proto.rawsend
pyskynet.ret = pyskynet_proto.ret

#################
# env set & get #
#################


def getenv(key):
    data = skynet_py_main.getlenv(key)
    if data is None:
        return None
    else:
        return foreign_seri.remoteunpack(data)[0]


def setenv(key, value):
    if skynet_py_main.self() != 0:
        assert (key is None) or (getenv(key) is None), "Can't setenv exist key : %s " % key
    msg_ptr, msg_size = foreign_seri.remotepack(value)
    newkey = skynet_py_main.setlenv(key, msg_ptr, msg_size)
    foreign_seri.trash(msg_ptr, msg_size)
    return newkey


def envs():
    key = None
    re = {}
    while True:
        key = skynet_py_main.nextenv(key)
        if(key is None):
            break
        else:
            re[key] = getenv(key)
    return re


###############
# service api #
###############
def newservice(service_name, *args):
    assert type(service_name) == str or type(service_name) == bytes, "newservice's name must be str or bytes"
    for arg in args:
        assert type(arg) == str or type(arg) == bytes, "newservice's arg must be str or bytes"
    return pyskynet_proto.call(".launcher", PTYPE_LUA, "LAUNCH", "snlua", service_name, *args)[0]


def uniqueservice(service_name, *args):
    assert type(service_name) == str or type(service_name) == bytes, "uniqueservice's name must be str or bytes"
    for arg in args:
        assert type(arg) == str or type(arg) == bytes, "uniqueservice's arg must be str or bytes"
    return pyskynet_proto.call(".service", PTYPE_LUA, "LAUNCH", service_name, *args)[0]


def scriptservice(scriptaddr_or_loadargs, *args):
    t1 = type(scriptaddr_or_loadargs)
    if t1 == str and scriptaddr_or_loadargs.startswith("0x"):
        scriptaddr = scriptaddr_or_loadargs
    elif t1 == bytes and scriptaddr_or_loadargs.startswith(b"0x"):
        scriptaddr = scriptaddr_or_loadargs
    elif t1 == str or t1 == bytes:
        scriptaddr = setenv(None, [scriptaddr_or_loadargs])
    elif t1 == list:
        scriptaddr = setenv(None, scriptaddr_or_loadargs)
    else:
        raise Exception("loadservice's first args must be str or bytes or list")
    return newservice("script_service", scriptaddr, *args)


class __CanvasService(object):
    def __init__(self, service):
        self.service = service

    def reset(self, *args):
        return pyskynet_proto.call(self.service, PTYPE_FOREIGN, "reset", *args)

    def render(self, *args):
        return pyskynet_proto.call(self.service, PTYPE_FOREIGN, "render", *args)

    def __del__(self):
        return pyskynet_proto.send(self.service, PTYPE_FOREIGN, "exit")


def canvas(script, name="unknowxml"):
    scriptaddr = setenv(None, script)
    return __CanvasService(newservice("canvas_service", scriptaddr, name))


def self():
    address = skynet_py_main.self()
    assert address > 0, "service pyholder not start "
    return address


def test(script, *args):
    import pyskynet.foreign
    return pyskynet.foreign.call(boot.boot_service, "run", script, *args)





import pyskynet

import pyskynet.boot
import pyskynet.proto as pyskynet_proto
from pyskynet.skynet_py_foreign_seri import remotepack, remoteunpack, luapack
import gevent
import gevent.event

clusterd = None


def __load_clusterd():
    global clusterd
    clusterd = pyskynet.uniqueservice("foreign_clusterd")


pyskynet.boot.init(__load_clusterd)

node_to_sender = {}
node_to_task_queue = {}


def __request_sender(task_queue, node):
    sender = pyskynet_proto.call(clusterd, "lua", "sender", node)[0]
    i = 0
    while i < len(task_queue):
        task = task_queue[i]
        # tuple for send task, AsyncResult for call task
        if type(task) == tuple:
            pyskynet_proto.send(sender, "lua", "push", task[0], remotepack(*task[1]))
        else:
            confirm = gevent.event.Event()
            task.set((confirm, sender))
            confirm.wait()
        i += 1
    node_to_task_queue.pop(node)
    node_to_sender[node] = sender


def __get_sender(node):
    sender = node_to_sender.get(node)
    if sender is None:
        task = gevent.event.AsyncResult()
        task_queue = node_to_task_queue.get(node)
        if task_queue is None:
            task_queue = [task]
            node_to_task_queue[node] = task_queue
            gevent.spawn(__request_sender, task_queue, node)
        else:
            task_queue.append(task)
        confirm, sender = task.get()
        confirm.set()
        return sender
    else:
        return sender


def call(node, address, *args):
    return remoteunpack(*pyskynet_proto.rawcall(
        __get_sender(node), "lua", *luapack("req", address, *remotepack(*args))))


def send(node, address, *args):
    sender = node_to_sender.get(node)
    if sender is None:
        node_to_task_queue[node].append((address, args))
    else:
        pyskynet_proto.send(sender, "lua", "push", address, remotepack(*args))


def open(port):
    assert 0 <= port <= 65535, "cluster.open's port must be in [0, 65535]"
    pyskynet_proto.call(clusterd, "lua", "listen", "0.0.0.0", port)




from distutils.command.build_ext import build_ext
from distutils.command.build import build
from setuptools import Extension, setup
import setuptools.command.install
import setuptools.command.develop

import os
import sys
import re


def list_path(path, suffix, exclude_files=[]):
    if path[-1] != "/":
        path += "/"
    re = []
    for f in os.listdir(path):
        if (f.rfind(suffix) == len(f) - len(suffix)) and (f not in exclude_files):
            re.append(path+f)
    return re

# SKYNET_3RD = "./skynet/3rd"

# JEMALLOC_STATICLIB = SKYNET_3RD + "/jemalloc/lib/libjemalloc_pic.a"
# JEMALLOC_INC = SKYNET_3RD + "/jemalloc/include/jemalloc"


LUA_PATH = "./skynet/3rd/lua"
SKYNET_SRC_PATH = "./skynet/skynet-src"

LIBRARIES = ["pthread", "m", "readline"]
MACROS = [("NOUSE_JEMALLOC", None), ("BUILD_FOR_PYSKYNET", None), ("__STDC_NO_ATOMICS__", None)]

TFLITE_LIB = None

if sys.platform == "linux":
    MACROS += [("LUA_USE_LINUX", None)]
    LIBRARIES += ["dl"]
    TFLITE_LIB = "3rd/nn_libs/tflite/lib/manylinux/libtensorflow-lite.a"
elif sys.platform == "darwin":
    MACROS += [("LUA_USE_MACOSX", None)]
    TFLITE_LIB = "3rd/nn_libs/tflite/lib/macosx/libtensorflow-lite.a"
else:
    TFLITE_LIB = "3rd/nn_libs/tflite/lib/win/tensorflow-lite.lib"
    raise Exception("no build config for platform %s" % sys.platform)

INCLUDE_DIRS = [SKYNET_SRC_PATH, LUA_PATH, "./src", "./src/c_src", "./skynet/lualib-src"]


def create_skynet_extensions():
    SKYNET_CSERVICES = ["snlua", "logger", "gate", "harbor"]
    ext_cservices = []
    for cservice in SKYNET_CSERVICES:
        ext = Extension('skynet.cservice.'+cservice,
            include_dirs=INCLUDE_DIRS,
            sources=['skynet/service-src/service_'+cservice+'.c'],
            define_macros=MACROS,
            extra_objects=[])
        ext_cservices.append(ext)
    clib_skynet_src = ["lua-skynet.c",
                "lua-seri.c",
                "lua-socket.c",
                "lua-mongo.c",
                "lua-netpack.c",
                "lua-memory.c",
                "lua-multicast.c",
                "lua-cluster.c",
                "lua-crypt.c",
                "lsha1.c",
                "lua-sharedata.c",
                "lua-stm.c",
                "lua-debugchannel.c",
                "lua-datasheet.c",
                "lua-sharetable.c"]
    ext_skynet = Extension('skynet.luaclib.skynet',
        include_dirs=INCLUDE_DIRS,
        define_macros=MACROS,
        sources=["skynet/lualib-src/" + s for s in clib_skynet_src],
        extra_objects=[])
    ext_lpeg = Extension('skynet.luaclib.lpeg',
        include_dirs=[LUA_PATH, "skynet/3rd/lpeg"],
        sources=list_path("skynet/3rd/lpeg", ".c"),
        define_macros=MACROS,
        extra_objects=[])
    ext_md5 = Extension('skynet.luaclib.md5',
        include_dirs=[LUA_PATH, "skynet/3rd/lua-md5"],
        sources=list_path("skynet/3rd/lua-md5", ".c"),
        define_macros=MACROS,
        extra_objects=[])
    ext_bson = Extension('skynet.luaclib.bson',
        include_dirs=[SKYNET_SRC_PATH, LUA_PATH, "skynet/lualib-src/lua-bson"],
        sources=["skynet/lualib-src/lua-bson.c"],
        define_macros=MACROS,
        extra_objects=[])
    return ext_cservices + [ext_skynet, ext_lpeg, ext_md5, ext_bson]

def create_skynet_extensions_ltls(ssl=False):
    tls_library_dirs=[]
    tls_include_dirs=[SKYNET_SRC_PATH, LUA_PATH, "skynet/lualib-src/ltls"]
    if ssl and type(ssl) == str:
        search_path = [ssl, "/usr", "/usr/local", "/usr/local/opt"]
    else:
        search_path = ["/usr", "/usr/local", "/usr/local/opt"]
    ssl_found = False
    for path in search_path:
        if os.path.isfile(path+"/include/openssl/ssl.h"):
            tls_library_dirs.append(path+"/lib")
            tls_include_dirs.append(path+"/include")
            ssl_found = True
            break
    if not ssl_found:
        if ssl:
            raise Exception("'openssl/ssl.h' not found")
        else:
            print("'openssl/ssl.h' not found, skynet/luaclib/ltls.so not installed")
        return []
    else:
        ext_ltls = Extension('skynet.luaclib.ltls',
            include_dirs=tls_include_dirs,
            library_dirs=tls_library_dirs,
            sources=["skynet/lualib-src/ltls.c"],
            libraries=["ssl"],
            define_macros=MACROS,
            extra_objects=[])
        return [ext_ltls]

def create_cython_extensions():
    ext_main = Extension('pyskynet.skynet_py_main',
        include_dirs=INCLUDE_DIRS,
        sources=['src/cy_src/skynet_py_main.pyx'] +
                list_path(SKYNET_SRC_PATH, ".c", ["skynet_main.c", "skynet_start.c", "skynet_env.c"]) +
                list_path("src/skynet_modify", ".c") +
                list_path("src/skynet_foreign", ".c", ["test.c"]) +
                list_path(LUA_PATH, ".c", ["lua.c", "luac.c"]),
        depends=['src/cy_src/skynet_py.pxd'],
        define_macros=MACROS,
        libraries=LIBRARIES,
        extra_objects=[])

    ext_seri = Extension('pyskynet.skynet_py_foreign_seri',
        include_dirs=INCLUDE_DIRS,
        sources=['src/cy_src/skynet_py_foreign_seri.pyx'],
        depends=['src/cy_src/skynet_py.pxd'],
        define_macros=MACROS,
        libraries=LIBRARIES)

    ext_mq = Extension('pyskynet.skynet_py_mq',
        include_dirs=INCLUDE_DIRS,
        sources=['src/cy_src/skynet_py_mq.pyx'],
        depends=['src/cy_src/skynet_py.pxd'],
        define_macros=MACROS,
        libraries=LIBRARIES)

    return [ext_main, ext_mq, ext_seri]


def create_lua_extensions():
    lua_service_pyholder = Extension('pyskynet.service.pyholder',
        sources=['src/c_src/service_pyholder.c'],
        include_dirs=INCLUDE_DIRS,
        define_macros=MACROS,
        libraries=LIBRARIES)
    lua_foreign_seri = Extension('pyskynet.lualib.pyskynet.foreign_seri',
        sources=['src/c_src/lua-foreign_seri.c'],
        include_dirs=INCLUDE_DIRS,
        define_macros=MACROS,
        libraries=LIBRARIES)
    lua_modify = Extension('pyskynet.lualib.pyskynet.modify',
        sources=['src/c_src/lua-modify.c'],
        include_dirs=INCLUDE_DIRS,
        define_macros=MACROS,
        libraries=LIBRARIES)
    lua_numsky = Extension('pyskynet.lualib.numsky',
        sources=list_path("src/c_src/numsky", ".cpp") +
                list_path("src/c_src/numsky/ndarray", ".cpp") +
                list_path("src/c_src/numsky/ufunc", ".cpp") +
                list_path("src/c_src/numsky/canvas", ".cpp") +
                list_path("src/c_src/numsky/tinygl", ".cpp") +
                list_path("3rd/TinyGL/tinygl", ".cpp"),
        include_dirs=INCLUDE_DIRS + ["3rd/rapidxml", "3rd/TinyGL"],
        define_macros=MACROS,
        extra_compile_args=['-std=c++11'],
        libraries=LIBRARIES)
    return [lua_service_pyholder, lua_foreign_seri, lua_modify, lua_numsky]


def create_3rd_extensions():
    lua_pb = Extension('pyskynet.lualib.pb',
        sources=["3rd/lua-protobuf/pb.c"],
        include_dirs=["3rd/lua-protobuf", LUA_PATH])
    lua_rapidjson = Extension('pyskynet.lualib.rapidjson',
        sources=list_path("3rd/lua-rapidjson/src", ".cpp"),
        extra_compile_args=["-std=c++11"],
        include_dirs=["3rd/lua-rapidjson/src", "3rd/lua-rapidjson/rapidjson/include", LUA_PATH, "3rd/"])
    lua_unqlite = Extension('pyskynet.lualib.unqlite',
        sources=["src/c_src/lua-unqlite.cpp", "3rd/unqlite/unqlite.c"],
        include_dirs=["3rd/unqlite/"]+INCLUDE_DIRS)
    return [lua_pb, lua_rapidjson, lua_unqlite]


def create_tflite_extensions():
    lua_tflite = Extension('pyskynet.lualib.tflite',
        sources=["src/c_src/lua-tflite.cpp", "src/c_src/numsky/lua-numsky.cpp"],
        include_dirs=INCLUDE_DIRS + ["3rd/nn_libs/tflite/tflite_cinclude"],
        define_macros=MACROS,
        extra_compile_args=['-std=c++11'],
        libraries=LIBRARIES,
        extra_objects=[TFLITE_LIB])
    return [lua_tflite]

install_opts = {
        "ssl":False,
        "tflite":False,
        }

class build_with_numpy_cython(build):
    def finalize_options(self):
        super().finalize_options()
        self.distribution.ext_modules=create_skynet_extensions() + create_cython_extensions() + create_lua_extensions() + create_3rd_extensions() + create_skynet_extensions_ltls(install_opts["ssl"])
        import numpy
        for extension in self.distribution.ext_modules:
            np_inc = numpy.get_include()
            if not (np_inc in extension.include_dirs):
                extension.include_dirs.append(np_inc)
        from Cython.Build import cythonize
        self.distribution.ext_modules = cythonize(self.distribution.ext_modules, language_level=3)


class build_ext_rename(build_ext):
    def get_ext_filename(self, ext_name):
        ext_name_last = ext_name.split(".")[-1]
        # cython library start with skynet_py
        if ext_name_last.find("skynet_py_") == 0:
            # for cython library
            return super().get_ext_filename(ext_name)
        else:
            # for lua library
            ext_path = ext_name.split('.')
            return os.path.join(*ext_path) + ".so"


#class build_ext_purec(build_ext):
#    def get_ext_filename(self, ext_name):
#        ext_path = ext_name.split('.')
#        return os.path.join(*ext_path) + ".so"


class CommandMixin(object):
    user_options = [
        ('ssl=', None, 'build with ssl'),
        ('tflite', None, 'build with tflite'),
    ]

    def initialize_options(self):
        super().initialize_options()
        # Initialize options
        self.ssl = False
        self.tflite = False

    def finalize_options(self):
        super().finalize_options()

    def run(self):
        # Use options
        install_opts["ssl"] = self.ssl
        install_opts["tflite"] = self.tflite
        super().run()


class InstallCommand(CommandMixin, setuptools.command.install.install):
    user_options = getattr(setuptools.command.install.install, 'user_options', []) + CommandMixin.user_options

class DevelopCommand(CommandMixin, setuptools.command.develop.develop):
    user_options = getattr(setuptools.command.develop.develop, 'user_options', []) + CommandMixin.user_options


def get_version():
    with open("pyskynet/__init__.py") as fo:
        data = fo.read()
        result = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', data)
        return result.group(1)


def main():
    setup(
            name="pyskynet",
            version=get_version(),
            author="cz",
            author_email="chenze.3057@bytedance.com",
            license='MIT',
            description="PySkynet is a library for using skynet in python.",
            ext_modules=[], # setted in build_with_numpy_cython
            cmdclass={"build_ext": build_ext_rename, "build": build_with_numpy_cython, "install":InstallCommand, "develop":DevelopCommand},
            packages=["pyskynet", "skynet"],
            package_data={
                "pyskynet": ["service/*",
                            "lualib/*",
                            "lualib/*/*"],
                "skynet": ["service/*",
                          "cservice/*",
                          "luaclib/*",
                          "lualib/*",
                          "lualib/*/*",
                          "lualib/*/*/*"],
            },
            zip_safe=False,
            entry_points={
                "console_scripts": [
                    "pyskynet=pyskynet.boot:main",
                ]
            },
            install_requires=[
                "cffi ~= 1.14.2",
                "gevent >= 20.6.0",
                "numpy",
            ],
            url='https://github.com/bytedance/pyskynet',
            setup_requires=["cython", "numpy"],
            python_requires='>=3.5',
        )


main()
